@@ -16,9 +16,10 @@ use strict;
 use vars qw($VERSION);
 use Image::ExifTool qw(:DataAccess :Utils);
 
-$VERSION = '1.26';
+$VERSION = '1.33';
 
 sub ProcessJpeg2000Box($$$);
+sub ProcessJUMD($$$);
 
 my %resolutionUnit = (
     -3 => 'km',
@@ -41,8 +42,25 @@ my %jp2Map = (
    'UUID-IPTC'   => 'JP2',
    'UUID-EXIF'   => 'JP2',
    'UUID-XMP'    => 'JP2',
-  # jp2h         => 'JP2',  (not yet functional)
-  # ICC_Profile  => 'jp2h', (not yet functional)
+    jp2h         => 'JP2',
+    colr         => 'jp2h',
+    ICC_Profile  => 'colr',
+    IFD1         => 'IFD0',
+    EXIF         => 'IFD0', # to write EXIF as a block
+    ExifIFD      => 'IFD0',
+    GPS          => 'IFD0',
+    SubIFD       => 'IFD0',
+    GlobParamIFD => 'IFD0',
+    PrintIM      => 'IFD0',
+    InteropIFD   => 'ExifIFD',
+    MakerNotes   => 'ExifIFD',
+);
+
+# map of where information is written in a JXL image
+my %jxlMap = (
+    IFD0         => 'Exif',
+    XMP          => 'XML',
+   'Exif'        => 'JP2',
     IFD1         => 'IFD0',
     EXIF         => 'IFD0', # to write EXIF as a block
     ExifIFD      => 'IFD0',
@@ -107,9 +125,13 @@ my %j2cMarker = (
     WRITE_PROC => \&ProcessJpeg2000Box,
     PREFERRED => 1, # always add these tags when writing
     NOTES => q{
-        The tags below are extracted from JPEG 2000 images, however ExifTool
-        currently writes only EXIF, IPTC and XMP tags in these images.
+        The tags below are found in JPEG 2000 images and the JUMBF metadata in JPEG
+        images, but not all of these are extracted.  Note that ExifTool currently
+        writes only EXIF, IPTC and XMP tags in Jpeg2000 images.
     },
+#
+# NOTE: ONLY TAGS WITH "Format" DEFINED ARE EXTRACTED!
+#
    'jP  ' => 'JP2Signature', # (ref 1)
    "jP\x1a\x1a" => 'JP2Signature', # (ref 2)
     prfl => 'Profile',
@@ -186,7 +208,11 @@ my %j2cMarker = (
     comp => 'Composition',
     copt => 'CompositionOptions',
     inst => 'InstructionSet',
-    asoc => 'Association',
+    asoc => {
+        Name => 'Association',
+        SubDirectory => { },
+    },
+        # (Association box may contain any other sub-box)
     nlst => 'NumberList',
     bfil => 'BinaryFilter',
     drep => 'DesiredReproductions',
@@ -195,13 +221,22 @@ my %j2cMarker = (
     chck => 'DigitalSignature',
     mp7b => 'MPEG7Binary',
     free => 'Free',
-    jp2c => 'ContiguousCodestream',
+    jp2c => [{
+        Name => 'ContiguousCodestream',
+        Condition => 'not $$self{jumd_level}',
+    },{
+        Name => 'PreviewImage',
+        Groups => { 2 => 'Preview' },
+        Format => 'undef',
+        Binary => 1,
+    }],
     jp2i => {
         Name => 'IntellectualProperty',
         SubDirectory => { TagTable => 'Image::ExifTool::XMP::Main' },
     },
-   'xml '=> {
+   'xml '=> [{
         Name => 'XML',
+        Condition => 'not $$self{IsJXL}',
         Writable => 'undef',
         Flags => [ 'Binary', 'Protected', 'BlockExtract' ],
         List => 1,
@@ -209,11 +244,18 @@ my %j2cMarker = (
             by default, the XML data in this tag is parsed using the ExifTool XMP module
             to to allow individual tags to be accessed when reading, but it may also be
             extracted as a block via the "XML" tag, which is also how this tag is
-            written and copied.  This is a List-type tag because multiple XML blocks may
-            exist
+            written and copied.  It may also be extracted as a block by setting the API
+            BlockExtract option.  This is a List-type tag because multiple XML blocks
+            may exist
         },
+        # (note: extracting as a block was broken in 11.04, and finally fixed in 12.14)
+        SubDirectory => { TagTable => 'Image::ExifTool::XMP::XML' },
+    },{
+        Name => 'XMP',
+        Notes => 'used for XMP in JPEG XL files',
+        # NOTE: the hacked code relies on this being at index 1 of the tagInfo list!
         SubDirectory => { TagTable => 'Image::ExifTool::XMP::Main' },
-    },
+    }],
     uuid => [
         {
             Name => 'UUID-EXIF',
@@ -297,6 +339,22 @@ my %j2cMarker = (
                 Start => '$valuePtr + 16',
             },
         },
+        {
+            Name => 'UUID-Signature',  # (seen in JUMB data of JPEG images)
+            # (may be able to remove this when JUMBF specification is finalized)
+            Condition => '$$valPt=~/^casg\x00\x11\x00\x10\x80\x00\x00\xaa\x00\x38\x9b\x71/',
+            Format => 'undef',
+            ValueConv => 'substr($val,16)',
+        },
+        {
+            Name => 'UUID-C2PAClaimSignature',  # (seen in incorrectly-formatted JUMB data of JPEG images)
+            # (may be able to remove this when JUMBF specification is finalized)
+            Condition => '$$valPt=~/^c2cs\x00\x11\x00\x10\x80\x00\x00\xaa\x00\x38\x9b\x71/',
+            SubDirectory => {
+                TagTable => 'Image::ExifTool::CBOR::Main',
+                Start => '$valuePtr + 16',
+            },
+        },
         {
             Name => 'UUID-Unknown',
         },
@@ -316,6 +374,79 @@ my %j2cMarker = (
             Name => 'URL',
             Format => 'string',
         },
+    # JUMBF boxes (ref https://github.com/thorfdbg/codestream-parser)
+    jumd => {
+        Name => 'JUMBFDescr',
+        SubDirectory => { TagTable => 'Image::ExifTool::Jpeg2000::JUMD' },
+    },
+    jumb => {
+        Name => 'JUMBFBox',
+        SubDirectory => {
+            TagTable => 'Image::ExifTool::Jpeg2000::Main',
+            ProcessProc => \&ProcessJUMB,
+        },
+    },
+    json => {
+        Name => 'JSONData',
+        Flags => [ 'Binary', 'Protected', 'BlockExtract' ],
+        Notes => q{
+            by default, data in this tag is parsed using the ExifTool JSON module to to
+            allow individual tags to be accessed when reading, but it may also be
+            extracted as a block via the "JSONData" tag or by setting the API
+            BlockExtract option
+        },
+        SubDirectory => { TagTable => 'Image::ExifTool::JSON::Main' },
+    },
+    cbor => {
+        Name => 'CBORData',
+        Flags => [ 'Binary', 'Protected' ],
+        SubDirectory => { TagTable => 'Image::ExifTool::CBOR::Main' },
+    },
+    bfdb => { # used in JUMBF (see  # (used when tag is renamed according to JUMDLabel)
+        Name => 'BinaryDataType',
+        Notes => 'JUMBF, MIME type and optional file name',
+        Format => 'undef',
+        # (ignore "toggles" byte and just extract MIME type and file name)
+        ValueConv => '$_=substr($val,1); s/\0+$//; s/\0/, /; $_',
+        JUMBF_Suffix => 'Type', # (used when tag is renamed according to JUMDLabel)
+    },
+    bidb => { # used in JUMBF
+        Name => 'BinaryData',
+        Notes => 'JUMBF',
+        Groups => { 2 => 'Preview' },
+        Format => 'undef',
+        Binary => 1,
+        JUMBF_Suffix => 'Data', # (used when tag is renamed according to JUMDLabel)
+    },
+    c2sh => { # used in JUMBF
+        Name => 'C2PASaltHash',
+        Format => 'undef',
+        ValueConv => 'unpack("H*",$val)',
+        JUMBF_Suffix => 'Salt', # (used when tag is renamed according to JUMDLabel)
+    },
+#
+# stuff seen in JPEG XL images:
+#
+  # jbrd - JPEG Bitstream Reconstruction Data (allows lossless conversion back to original JPG)
+    jxlc => {
+        Name => 'JXLCodestream',
+        Format => 'undef',
+        Notes => q{
+            Codestream in JPEG XL image.  Currently processed only to determine
+            ImageSize
+        },
+        RawConv => 'Image::ExifTool::Jpeg2000::ProcessJXLCodestream($self,\$val); undef',
+    },
+    Exif => {
+        Name => 'EXIF',
+        SubDirectory => {
+            TagTable => 'Image::ExifTool::Exif::Main',
+            ProcessProc => \&Image::ExifTool::ProcessTIFF,
+            WriteProc => \&Image::ExifTool::WriteTIFF,
+            DirName => 'EXIF',
+            Start => '$valuePtr + 4',
+        },
+    },
 );
 
 %Image::ExifTool::Jpeg2000::ImageHeader = (
@@ -370,6 +501,7 @@ my %j2cMarker = (
             'jp2 ' => 'JPEG 2000 Image (.JP2)',           # image/jp2
             'jpm ' => 'JPEG 2000 Compound Image (.JPM)',  # image/jpm
             'jpx ' => 'JPEG 2000 with extensions (.JPX)', # image/jpx
+            'jxl ' => 'JPEG XL Image (.JXL)',             # image/jxl
         },
     },
     1 => {
@@ -435,11 +567,34 @@ my %j2cMarker = (
 
 %Image::ExifTool::Jpeg2000::ColorSpec = (
     PROCESS_PROC => \&Image::ExifTool::ProcessBinaryData,
+    WRITE_PROC => \&Image::ExifTool::WriteBinaryData, # (we don't actually call this)
     GROUPS => { 2 => 'Image' },
     FORMAT => 'int8s',
+    WRITABLE => 1,
+    # (Note: 'colr' is not a real group, but is used as a hack to write the
+    #  necessary colr box.  This hack necessitated another hack in TagInfoXML.pm
+    #  to avoid reporting this fake group in the XML output)
+    WRITE_GROUP => 'colr',
+    DATAMEMBER => [ 0 ],
+    IS_SUBDIR => [ 3 ],
+    NOTES => q{
+        The table below contains tags in the color specification (colr) box.  This
+        box may be rewritten by writing either ICC_Profile, ColorSpace or
+        ColorSpecData.  When writing, any existing colr boxes are replaced with the
+        newly created colr box.
+
+        B<NOTE>: Care must be taken when writing this color specification because
+        writing a specification that is incompatible with the image data may make
+        the image undisplayable.
+    },
     0 => {
         Name => 'ColorSpecMethod',
         RawConv => '$$self{ColorSpecMethod} = $val',
+        Protected => 1,
+        Notes => q{
+            default for writing is 2 when writing ICC_Profile, 1 when writing
+            ColorSpace, or 4 when writing ColorSpecData
+        },
         PrintConv => {
             1 => 'Enumerated',
             2 => 'Restricted ICC',
@@ -447,9 +602,15 @@ my %j2cMarker = (
             4 => 'Vendor Color',
         },
     },
-    1 => 'ColorSpecPrecedence',
+    1 => {
+        Name => 'ColorSpecPrecedence',
+        Notes => 'default for writing is 0',
+        Protected => 1,
+    },
     2 => {
         Name => 'ColorSpecApproximation',
+        Notes => 'default for writing is 0',
+        Protected => 1,
         PrintConv => {
             0 => 'Not Specified',
             1 => 'Accurate',
@@ -474,6 +635,7 @@ my %j2cMarker = (
             Name => 'ColorSpace',
             Condition => '$$self{ColorSpecMethod} == 1',
             Format => 'int32u',
+            Protected => 1,
             PrintConv => { # ref 15444-2 2002-05-15
                 0 => 'Bi-level',
                 1 => 'YCbCr(1)',
@@ -503,11 +665,136 @@ my %j2cMarker = (
         {
             Name => 'ColorSpecData',
             Format => 'undef[$size-3]',
+            Writable => 'undef',
+            Protected => 1,
             Binary => 1,
         },
     ],
 );
 
+# JUMBF description box
+%Image::ExifTool::Jpeg2000::JUMD = (
+    PROCESS_PROC => \&ProcessJUMD,
+    GROUPS => { 0 => 'JUMBF', 1 => 'JUMBF', 2 => 'Image' },
+    NOTES => 'Information extracted from the JUMBF description box.',
+    'type' => {
+        Name => 'JUMDType',
+        ValueConv => 'unpack "H*", $val',
+        PrintConv => q{
+            my @a = $val =~ /^(\w{8})(\w{4})(\w{4})(\w{16})$/;
+            return $val unless @a;
+            my $ascii = pack 'H*', $a[0];
+            $a[0] = "($ascii)" if $ascii =~ /^[a-zA-Z0-9]{4}$/;
+            return join '-', @a;
+        },
+        # seen:
+        # cacb/cast/caas/cacl/casg/json-00110010800000aa00389b71
+        # 6579d6fbdba2446bb2ac1b82feeb89d1 - JPEG image
+    },
+    'label' => { Name => 'JUMDLabel' },
+    'toggles' => {
+        Name => 'JUMDToggles',
+        Unknown => 1,
+        PrintConv => { BITMASK => {
+            0 => 'Requestable',
+            1 => 'Label',
+            2 => 'ID',
+            3 => 'Signature',
+        }},
+    },
+    'id'    => { Name => 'JUMDID', Description => 'JUMD ID' },
+    'sig'   => { Name => 'JUMDSignature', PrintConv => 'unpack "H*", $val' },
+);
+
+#------------------------------------------------------------------------------
+# Read JUMBF box to keep track of sub-document numbers
+# Inputs: 0) ExifTool ref, 1) dirInfo ref, 2) tag table ref
+# Returns: 1 on success
+sub ProcessJUMB($$$)
+{
+    my ($et, $dirInfo, $tagTablePtr) = @_;
+    if ($$et{jumd_level}) {
+        ++$$et{jumd_level}[-1]; # increment current sub-document number
+    } else {
+        $$et{jumd_level} = [ ++$$et{DOC_COUNT} ]; # new top-level sub-document
+        $$et{SET_GROUP0} = 'JUMBF';
+    }
+    $$et{DOC_NUM} = join '-', @{$$et{jumd_level}};
+    push @{$$et{jumd_level}}, 0;
+    ProcessJpeg2000Box($et, $dirInfo, $tagTablePtr);
+    delete $$et{DOC_NUM};
+    delete $$et{JUMBFLabel};
+    pop @{$$et{jumd_level}};
+    if (@{$$et{jumd_level}} < 2) {
+        delete $$et{jumd_level};
+        delete $$et{SET_GROUP0};
+    }
+    return 1;
+}
+
+#------------------------------------------------------------------------------
+# Read JUMBF description box (ref https://github.com/thorfdbg/codestream-parser)
+# Inputs: 0) ExifTool ref, 1) dirInfo ref, 2) tag table ref
+# Returns: 1 on success
+sub ProcessJUMD($$$)
+{
+    my ($et, $dirInfo, $tagTablePtr) = @_;
+    my $dataPt = $$dirInfo{DataPt};
+    my $pos    = $$dirInfo{DirStart};
+    my $end    = $pos + $$dirInfo{DirLen};
+    $et->VerboseDir('JUMD', 0, $end-$pos);
+    delete $$et{JUMBFLabel};
+    $$dirInfo{DirLen} < 17 and $et->Warn('Truncated JUMD directory'), return 0;
+    my $type = substr($$dataPt, $pos, 4);
+    $et->HandleTag($tagTablePtr, 'type', substr($$dataPt, $pos, 16));
+    $pos += 16;
+    my $flags = Get8u($dataPt, $pos++);
+    $et->HandleTag($tagTablePtr, 'toggles', $flags);
+    if ($flags & 0x02) {    # label exists?
+        pos($$dataPt) = $pos;
+        $$dataPt =~ /\0/g or $et->Warn('Missing JUMD label terminator'), return 0;
+        my $len = pos($$dataPt) - $pos;
+        my $name = substr($$dataPt, $pos, $len);
+        $et->HandleTag($tagTablePtr, 'label', $name);
+        $pos += $len;
+        if ($len) {
+            $name =~ s/[^-_a-zA-Z0-9]([a-z])/\U$1/g; # capitalize characters after illegal characters
+            $name =~ tr/-_a-zA-Z0-9//dc;    # remove other illegal characters
+            $name =~ s/__/_/;               # collapse double underlines
+            $name = ucfirst $name;          # capitalize first letter
+            $name = "Tag$name" if length($name) < 2; # must at least 2 characters long
+            $$et{JUMBFLabel} = $name;
+        }
+    }
+    if ($flags & 0x04) {    # ID exists?
+        $pos + 4 > $end and $et->Warn('Missing JUMD ID'), return 0;
+        $et->HandleTag($tagTablePtr, 'id', Get32u($dataPt, $pos));
+        $pos += 4;
+    }
+    if ($flags & 0x08) {    # signature exists?
+        $pos + 32 > $end and $et->Warn('Missing JUMD signature'), return 0;
+        $et->HandleTag($tagTablePtr, 'sig', substr($$dataPt, $pos, 32));
+        $pos += 32;
+    }
+    my $more = $end - $pos;
+    if ($more) {
+        # (may find c2sh box hiding after JUMD record)
+        if ($more >= 8) {
+            my %dirInfo = (
+                DataPt   => $dataPt,
+                DataLen  => $$dirInfo{DataLen},
+                DirStart => $pos,
+                DirLen   => $more,
+                DirName  => 'JUMDPrivate',
+            );
+            $et->ProcessDirectory(\%dirInfo, GetTagTable('Image::ExifTool::Jpeg2000::Main'));
+        } else {
+            $et->Warn("Extra data in JUMD box $more bytes)", 1);
+        }
+    }
+    return 1;
+}
+
 #------------------------------------------------------------------------------
 # Create new JPEG 2000 boxes when writing
 # (Currently only supports adding top-level Writable JPEG2000 tags and certain UUID boxes)
@@ -537,8 +824,29 @@ sub CreateNewBoxes($$)
             $et->VerboseValue("+ Jpeg2000:$$tagInfo{Name}", $val);
         }
     }
-    # add UUID boxes
+    # add UUID boxes (and/or JXL Exif/XML boxes)
     foreach $dirName (sort keys %$addDirs) {
+        # handle JPEG XL XMP and EXIF
+        if ($dirName eq 'XML' or $dirName eq 'Exif') {
+            my ($tag, $dir) = $dirName eq 'XML' ? ('xml ', 'XMP') : ('Exif', 'EXIF');
+            my $tagInfo = $Image::ExifTool::Jpeg2000::Main{$tag};
+            $tagInfo = $$tagInfo[1] if ref $tagInfo eq 'ARRAY'; # (hack for stupid JXL XMP)
+            my $subdir = $$tagInfo{SubDirectory};
+            my $tagTable = GetTagTable($$subdir{TagTable});
+            $tagTable = GetTagTable('Image::ExifTool::XMP::Main') if $dir eq 'XMP';
+            my %dirInfo = (
+                DirName => $dir,
+                Parent => 'JP2',
+            );
+            my $newdir = $et->WriteDirectory(\%dirInfo, $tagTable, $$subdir{WriteProc});
+            if (defined $newdir and length $newdir) {
+                # not sure why, but EXIF box is padded with leading 0's in my sample
+                my $pad = $dirName eq 'Exif' ? "\0\0\0\0" : '';
+                my $boxhdr = pack('N', length($newdir) + length($pad) + 8) . $tag;
+                Write($outfile, $boxhdr, $pad, $newdir) or return 0;
+                next;
+            }
+        }
         next unless $uuid{$dirName};
         my $tagInfo;
         foreach $tagInfo (@{$Image::ExifTool::Jpeg2000::Main{uuid}}) {
@@ -563,6 +871,48 @@ sub CreateNewBoxes($$)
     return 1;
 }
 
+#------------------------------------------------------------------------------
+# Create Color Specification Box
+# Inputs: 0) ExifTool object ref, 1) Output file or scalar ref
+# Returns: 1 on success
+sub CreateColorSpec($$)
+{
+    my ($et, $outfile) = @_;
+    my $meth   = $et->GetNewValue('Jpeg2000:ColorSpecMethod');
+    my $prec   = $et->GetNewValue('Jpeg2000:ColorSpecPrecedence') || 0;
+    my $approx = $et->GetNewValue('Jpeg2000:ColorSpecApproximation') || 0;
+    my $icc    = $et->GetNewValue('ICC_Profile');
+    my $space  = $et->GetNewValue('Jpeg2000:ColorSpace');
+    my $cdata  = $et->GetNewValue('Jpeg2000:ColorSpecData');
+    unless ($meth) {
+        if ($icc) {
+            $meth = 2;
+        } elsif (defined $space) {
+            $meth = 1;
+        } elsif (defined $cdata) {
+            $meth = 4;
+        } else {
+            $et->Warn('Color space not defined'), return 0;
+        }
+    }
+    if ($meth eq '1') {
+        defined $space or $et->Warn('Must specify ColorSpace'), return 0;
+        $cdata = pack('N', $space);
+    } elsif ($meth eq '2' or $meth eq '3') {
+        defined $icc or $et->Warn('Must specify ICC_Profile'), return 0;
+        $cdata = $icc;
+    } elsif ($meth eq '4') {
+        defined $cdata or $et->Warn('Must specify ColorSpecData'), return 0;
+    } else {
+        $et->Warn('Unknown ColorSpecMethod'), return 0;
+    }
+    my $boxhdr = pack('N', length($cdata) + 11) . 'colr';
+    Write($outfile, $boxhdr, pack('CCC',$meth,$prec,$approx), $cdata) or return 0;
+    ++$$et{CHANGED};
+    $et->VPrint(1, "    + Jpeg2000:ColorSpec\n");
+    return 1;
+}
+
 #------------------------------------------------------------------------------
 # Process JPEG 2000 box
 # Inputs: 0) ExifTool object reference, 1) dirInfo reference, 2) Pointer to tag table
@@ -573,14 +923,14 @@ sub ProcessJpeg2000Box($$$)
     my ($et, $dirInfo, $tagTablePtr) = @_;
     my $dataPt = $$dirInfo{DataPt};
     my $dataLen = $$dirInfo{DataLen};
-    my $dataPos = $$dirInfo{DataPos};
+    my $dataPos = $$dirInfo{DataPos} || 0;
     my $dirLen = $$dirInfo{DirLen} || 0;
     my $dirStart = $$dirInfo{DirStart} || 0;
     my $base = $$dirInfo{Base} || 0;
     my $raf = $$dirInfo{RAF};
     my $outfile = $$dirInfo{OutFile};
     my $dirEnd = $dirStart + $dirLen;
-    my ($err, $outBuff, $verbose);
+    my ($err, $outBuff, $verbose, $doColour);
 
     if ($outfile) {
         unless ($raf) {
@@ -588,13 +938,19 @@ sub ProcessJpeg2000Box($$$)
             $outBuff = '';
             $outfile = \$outBuff;
         }
+        # determine if we will be writing colr box
+        if ($$dirInfo{DirName} and $$dirInfo{DirName} eq 'JP2Header') {
+            $doColour = 2 if defined $et->GetNewValue('ColorSpecMethod') or $et->GetNewValue('ICC_Profile') or
+                             defined $et->GetNewValue('ColorSpecPrecedence') or defined $et->GetNewValue('ColorSpace') or
+                             defined $et->GetNewValue('ColorSpecApproximation') or defined $et->GetNewValue('ColorSpecData');
+        }
     } else {
         # (must not set verbose flag when writing!)
         $verbose = $$et{OPTIONS}{Verbose};
         $et->VerboseDir($$dirInfo{DirName}) if $verbose;
     }
     # loop through all contained boxes
-    my ($pos, $boxLen);
+    my ($pos, $boxLen, $lastBox);
     for ($pos=$dirStart; ; $pos+=$boxLen) {
         my ($boxID, $buff, $valuePtr);
         my $hdrLen = 8;     # the box header length
@@ -603,9 +959,7 @@ sub ProcessJpeg2000Box($$$)
             my $n = $raf->Read($buff,$hdrLen);
             unless ($n == $hdrLen) {
                 $n and $err = '', last;
-                if ($outfile) {
-                    CreateNewBoxes($et, $outfile) or $err = 1;
-                }
+                CreateNewBoxes($et, $outfile) or $err = 1 if $outfile;
                 last;
             }
             $dataPt = \$buff;
@@ -617,6 +971,17 @@ sub ProcessJpeg2000Box($$$)
         }
         $boxLen = unpack("x$pos N",$$dataPt);   # (length includes header and data)
         $boxID = substr($$dataPt, $pos+4, 4);
+        # remove old colr boxes if necessary
+        if ($doColour and $boxID eq 'colr') {
+            if ($doColour == 1) { # did we successfully write the new colr box?
+                $et->VPrint(1,"    - Jpeg2000:ColorSpec\n");
+                ++$$et{CHANGED};
+                next;
+            }
+            $et->Warn('Out-of-order colr box encountered');
+            undef $doColour;
+        }
+        $lastBox = $boxID;
         $pos += $hdrLen;                # move to end of box header
         if ($boxLen == 1) {
             # box header contains an additional 8-byte integer for length
@@ -644,7 +1009,7 @@ sub ProcessJpeg2000Box($$$)
                     }
                 } elsif ($verbose) {
                     my $msg = sprintf("offset 0x%.4x to end of file", $dataPos + $base + $pos);
-                    $et->VPrint(0, "$$et{INDENT}- Tag '$boxID' ($msg)\n");
+                    $et->VPrint(0, "$$et{INDENT}- Tag '${boxID}' ($msg)\n");
                 }
                 last;   # (ignore the rest of the file when reading)
             }
@@ -710,6 +1075,14 @@ sub ProcessJpeg2000Box($$$)
                 }
             }
         }
+        # create new tag for JUMBF data values with name corresponding to JUMBFLabel
+        if ($tagInfo and $$et{JUMBFLabel} and (not $$tagInfo{SubDirectory} or $$tagInfo{BlockExtract})) {
+            $tagInfo = { %$tagInfo, Name => $$et{JUMBFLabel} . ($$tagInfo{JUMBF_Suffix} || '') };
+            delete $$tagInfo{Description};
+            AddTagToTable($tagTablePtr, '_JUMBF_' . $$et{JUMBFLabel}, $tagInfo);
+            delete $$tagInfo{Protected}; # (must do this so -j -b returns JUMBF binary data)
+            $$tagInfo{TagID} = $boxID;
+        }
         if ($verbose) {
             $et->VerboseInfo($boxID, $tagInfo,
                 Table  => $tagTablePtr,
@@ -747,8 +1120,10 @@ sub ProcessJpeg2000Box($$$)
                 # remove this directory from our create list
                 delete $$et{AddJp2Dirs}{$$tagInfo{Name}};
                 my $newdir;
-                # only edit writable UUID boxes
-                if ($uuid) {
+                # only edit writable UUID, Exif and jp2h boxes
+                if ($uuid or $boxID eq 'Exif' or ($boxID eq 'xml ' and $$et{IsJXL}) or
+                    ($boxID eq 'jp2h' and $$et{EDIT_DIRS}{jp2h}))
+                {
                     $newdir = $et->WriteDirectory(\%subdirInfo, $subTable, $$subdir{WriteProc});
                     next if defined $newdir and not length $newdir; # next if deleting the box
                 } elsif (defined $uuid) {
@@ -760,6 +1135,11 @@ sub ProcessJpeg2000Box($$$)
                 my $boxhdr = pack('N', length($newdir) + 8 + $prefixLen) . $boxID;
                 $boxhdr .= substr($$dataPt, $valuePtr, $prefixLen) if $prefixLen;
                 Write($outfile, $boxhdr, $newdir) or $err = 1;
+                # write new colr box immediately after ihdr
+                if ($doColour and $boxID eq 'ihdr') {
+                    # (shouldn't be multiple ihdr boxes, but just in case, write only 1)
+                    $doColour = $doColour==2 ? CreateColorSpec($et, $outfile) : 0;
+                }
             } else {
                 # extract as a block if specified
                 $subdirInfo{BlockInfo} = $tagInfo if $$tagInfo{BlockExtract};
@@ -798,6 +1178,68 @@ sub ProcessJpeg2000Box($$$)
     return 1;
 }
 
+#------------------------------------------------------------------------------
+# Return bits from a bitstream object
+# Inputs: 0) array ref, 1) number of bits
+# Returns: specified number of bits as an integer, and shifts input bitstream
+sub GetBits($$)
+{
+    my ($a, $n) = @_;
+    my $v = 0;
+    my $bit = 1;
+    my $i;
+    while ($n--) {
+        for ($i=0; $i<@$a; ++$i) {
+            # consume bits LSB first
+            my $set = $$a[$i] & 1;
+            $$a[$i] >>= 1;
+            if ($i) {
+                $$a[$i-1] |= 0x80 if $set;
+            } else {
+                $v |= $bit if $set;
+                $bit <<= 1;
+            }
+        }
+    }
+    return $v;
+}
+
+#------------------------------------------------------------------------------
+# Extract parameters from JPEG XL codestream [unverified!]
+# Inputs: 0) ExifTool ref, 1) codestream ref
+# Returns: 1
+sub ProcessJXLCodestream($$)
+{
+    my ($et, $dataPt) = @_;
+    # add padding if necessary to avoid unpacking past end of data
+    if (length $$dataPt < 14) {
+        my $tmp = $$dataPt . ("\0" x 14);
+        $dataPt = \$tmp;
+    }
+    my @a = unpack 'x2C12', $$dataPt;
+    my ($x, $y);
+    my $small = GetBits(\@a, 1);
+    if ($small) {
+        $y = (GetBits(\@a, 5) + 1) * 8;
+    } else {
+        $y = GetBits(\@a, [9, 13, 18, 30]->[GetBits(\@a, 2)]) + 1;
+    }
+    my $ratio = GetBits(\@a, 3);
+    if ($ratio == 0) {
+        if ($small) {
+            $x = (GetBits(\@a, 5) + 1) * 8;;
+        } else {
+            $x = GetBits(\@a, [9, 13, 18, 30]->[GetBits(\@a, 2)]) + 1;
+        }
+    } else {
+        my $r = [[1,1],[12,10],[4,3],[3,2],[16,9],[5,4],[2,1]]->[$ratio-1];
+        $x = int($y * $$r[0] / $$r[1]);
+    }
+    $et->FoundTag(ImageWidth => $x);
+    $et->FoundTag(ImageHeight => $y);
+    return 1;
+}
+
 #------------------------------------------------------------------------------
 # Read/write meta information from a JPEG 2000 image
 # Inputs: 0) ExifTool object reference, 1) dirInfo reference
@@ -812,8 +1254,9 @@ sub ProcessJP2($$)
 
     # check to be sure this is a valid JPG2000 file
     return 0 unless $raf->Read($hdr,12) == 12;
-    unless ($hdr eq "\x00\x00\x00\x0cjP  \x0d\x0a\x87\x0a" or     # (ref 1)
-            $hdr eq "\x00\x00\x00\x0cjP\x1a\x1a\x0d\x0a\x87\x0a") # (ref 2)
+    unless ($hdr eq "\0\0\0\x0cjP  \x0d\x0a\x87\x0a" or     # (ref 1)
+            $hdr eq "\0\0\0\x0cjP\x1a\x1a\x0d\x0a\x87\x0a" or # (ref 2)
+            $$et{IsJXL})
     {
         return 0 unless $hdr =~ /^\xff\x4f\xff\x51\0/;  # check for JP2 codestream format
         if ($outfile) {
@@ -830,17 +1273,23 @@ sub ProcessJP2($$)
     }
     if ($outfile) {
         Write($outfile, $hdr) or return -1;
-        $et->InitWriteDirs(\%jp2Map);
+        if ($$et{IsJXL}) {
+            $et->InitWriteDirs(\%jxlMap);
+            $$et{AddJp2Tags} = { }; # (don't add JP2 tags in JXL files)
+        } else {
+            $et->InitWriteDirs(\%jp2Map);
+            $$et{AddJp2Tags} = $et->GetNewTagInfoHash(\%Image::ExifTool::Jpeg2000::Main);
+        }
         # save list of directories to create
-        my %addDirs = %{$$et{ADD_DIRS}};
+        my %addDirs = %{$$et{ADD_DIRS}}; # (make a copy)
         $$et{AddJp2Dirs} = \%addDirs;
-        $$et{AddJp2Tags} = $et->GetNewTagInfoHash(\%Image::ExifTool::Jpeg2000::Main);
     } else {
         my ($buff, $fileType);
         # recognize JPX and JPM as unique types of JP2
         if ($raf->Read($buff, 12) == 12 and $buff =~ /^.{4}ftyp(.{4})/s) {
             $fileType = 'JPX' if $1 eq 'jpx ';
             $fileType = 'JPM' if $1 eq 'jpm ';
+            $fileType = 'JXL' if $1 eq 'jxl ';
         }
         $raf->Seek(-length($buff), 1) if defined $buff;
         $et->SetFileType($fileType);
@@ -855,6 +1304,59 @@ sub ProcessJP2($$)
     return $et->ProcessDirectory(\%dirInfo, $tagTablePtr);
 }
 
+#------------------------------------------------------------------------------
+# Read meta information from a JPEG XL image
+# Inputs: 0) ExifTool object reference, 1) dirInfo reference
+# Returns: 1 on success, 0 if this wasn't a valid JPEG XL file, -1 on write error
+sub ProcessJXL($$)
+{
+    my ($et, $dirInfo) = @_;
+    my $raf = $$dirInfo{RAF};
+    my $outfile = $$dirInfo{OutFile};
+    my ($hdr, $buff);
+
+    return 0 unless $raf->Read($hdr,12) == 12;
+    if ($hdr eq "\0\0\0\x0cJXL \x0d\x0a\x87\x0a") {
+        # JPEG XL in ISO BMFF container
+        $$et{IsJXL} = 1;
+    } elsif ($hdr =~ /^\xff\x0a/) {
+        # JPEG XL codestream
+        if ($outfile) {
+            if ($$et{OPTIONS}{IgnoreMinorErrors}) {
+                $et->Warn('Wrapped JXL codestream in ISO BMFF container');
+            } else {
+                $et->Error('Will wrap JXL codestream in ISO BMFF container for writing',1);
+                return 0;
+            }
+            $$et{IsJXL} = 2;
+            my $buff = "\0\0\0\x0cJXL \x0d\x0a\x87\x0a\0\0\0\x14ftypjxl \0\0\0\0jxl ";
+            # add metadata to empty ISO BMFF container
+            $$dirInfo{RAF} = new File::RandomAccess(\$buff);
+        } else {
+            $et->SetFileType('JXL Codestream','image/jxl', 'jxl');
+            return ProcessJXLCodestream($et, \$hdr);
+        }
+    } else {
+        return 0;
+    }
+    $raf->Seek(0,0) or $et->Error('Seek error'), return 0;
+
+    my $success = ProcessJP2($et, $dirInfo);
+
+    if ($outfile and $success > 0 and $$et{IsJXL} == 2) {
+        # attach the JXL codestream box to the ISO BMFF file
+        $raf->Seek(0,2) or return -1;
+        my $size = $raf->Tell();
+        $raf->Seek(0,0) or return -1;
+        SetByteOrder('MM');
+        Write($outfile, Set32u($size + 8), 'jxlc') or return -1;
+        while ($raf->Read($buff, 65536)) {
+            Write($outfile, $buff) or return -1;
+        }
+    }
+    return $success;
+}
+
 1;  # end
 
 __END__
@@ -874,7 +1376,7 @@ files.
 
 =head1 AUTHOR
 
-Copyright 2003-2016, Phil Harvey (phil at owl.phy.queensu.ca)
+Copyright 2003-2022, Phil Harvey (philharvey66 at gmail.com)
 
 This library is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
